import sys
import os
import shutil
import subprocess
import glob 
import re
import json

from importlib import resources
from pdbtools import pdb_mkensemble
from nerrds.functions import check_quiet_print

natsort = lambda s: [int(t) if t.isdigit() else t.lower() for t in re.split('(\\d+)', s)] # used throughout to loop through files in natural order


def refine(home_dir, job_id, pdb, cpu, tleap, sander, sander_MPI, cpptraj, quiet=False):

    wd = home_dir+'/'+job_id+'/refined/'+os.path.splitext(os.path.basename(pdb))[0] # make dir to run refinement in
    os.makedirs(wd,exist_ok=True)

    with resources.path("nerrds.scripts.AMBER", "leap.in") as f:
        leap_script = f
    with resources.path("nerrds.scripts.AMBER", "min.in") as f:
        min_script = f
    with resources.path("nerrds.scripts.AMBER", "min_solv.in") as f:
        min_solv_script = f
    with resources.path("nerrds.scripts.AMBER", "extract_pdb.cpptraj") as f:
        cpptraj_script = f

    shutil.copyfile(leap_script, wd+"/leap.in") # copy various refinement scripts into wd
    shutil.copyfile(min_script, wd+"/min.in")  
    shutil.copyfile(min_solv_script, wd+"/min_solv.in")
    shutil.copyfile(pdb, wd+'/prot_for_min.tmp') # pdb is renamed to generic name referred to in refinement scripts, will be renamed after

    os.chdir(wd)

    out = open("prot_for_min.pdb","w")
    for line in open("prot_for_min.tmp","r"):
        if "OT" not in line:
            out.write(line)
    out.close()
    os.remove("prot_for_min.tmp")

    run_tleap = subprocess.run([tleap, "-f", "leap.in"],stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'))

    if int(cpu)==1:  
        check_quiet_print(quiet,home_dir+'/'+job_id,"refining "+pdb+" with 1 cpu using sander")
        run_amber = subprocess.run([sander, "-i", "min_solv.in", "-o", "min_solv.out", "-p", "prot_solv.parm7", "-c", "prot_solv.rst7", "-r", "prot_min_solv.rst7", "-x", "prot_min_solv.nc", "-ref", "prot_solv.rst7", "-O"],stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'))
        run_amber = subprocess.run([sander, "-i", "min.in", "-o", "min.out", "-p", "prot_solv.parm7", "-c", "prot_min_solv.rst7", "-r", "prot_min.rst7", "-x", "prot_min.nc", "-O"],stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'))
    elif int(cpu) > 1:
        check_quiet_print(quiet,home_dir+'/'+job_id,"refining "+pdb+" with "+str(cpu)+" cpus using sander.MPI")
        run_amber = subprocess.run(["mpirun", "-np", str(cpu), sander_MPI, "-i", "min_solv.in", "-o", "min_solv.out", "-p", "prot_solv.parm7", "-c", "prot_solv.rst7", "-r", "prot_min_solv.rst7", "-x", "prot_min_solv.nc", "-ref", "prot_solv.rst7", "-O"],stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'))
        run_amber = subprocess.run(["mpirun", "-np", str(cpu), sander_MPI, "-i", "min.in", "-o", "min.out", "-p", "prot_solv.parm7", "-c", "prot_min_solv.rst7", "-r", "prot_min.rst7", "-x", "prot_min.nc", "-O"],stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'))

    if os.path.exists("prot_min.nc"):
        run_cpptraj = subprocess.run([cpptraj, "-i", cpptraj_script],stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'))
        if os.path.exists("prot_min.pdb"): # sometimes cpptraj fails
            shutil.copyfile('prot_min.pdb',wd.split('/')[-1]+'_refined.pdb')
            os.remove('prot_min.pdb') 

    
    # remove traj - way too much space
    to_remove = glob.glob("*.nc")
    for t in to_remove:
        os.remove(t)
    to_remove = glob.glob("*.parm7")
    for t in to_remove:
        os.remove(t)
    to_remove = glob.glob("*.rst7")
    for t in to_remove:
        os.remove(t)
    to_remove = glob.glob("*.in")
    for t in to_remove:
        os.remove(t)

    os.remove("prot_for_min.pdb")
    os.chdir(home_dir)


def combine(home_dir, job_id, pdb_id,quiet=False):
    os.chdir(home_dir)

    # remove traj data for failures

    to_remove = glob.glob(home_dir+'/'+job_id+'/refined/anm*/*.nc')
    for t in to_remove:
        os.remove(t)
    to_remove = glob.glob(home_dir+'/'+job_id+'/refined/anm*/*.parm7')
    for t in to_remove:
        os.remove(t)
    to_remove = glob.glob(home_dir+'/'+job_id+'/refined/anm*/*.rst7')
    for t in to_remove:
        os.remove(t)
    to_remove = glob.glob(home_dir+'/'+job_id+'/refined/anm*/*.in')
    for t in to_remove:
        os.remove(t)


    mkensemble_string = []
    refined_pdbs = 0
    for pdb in sorted(glob.glob(home_dir+'/'+job_id+"/refined/anm_"+pdb_id+"_*/anm_"+pdb_id+"_*_refined.pdb"),key=natsort):
        fail=0
        for i in open(pdb,'r').readlines():  # remove the structures which blew up i.e. have coords with "***"
            if "*" in i:
                fail=1
        if fail == 0:
            mkensemble_string.append(pdb)
            refined_pdbs +=1

    new_pdb = pdb_mkensemble.run(mkensemble_string)

    out = open(home_dir+"/"+job_id+"/refined/anm_"+pdb_id+"_refined.tmp",'w')
    for line in enumerate(new_pdb):
        out.write(line[1])
    out.close()

    # rename HIE, HID to HIS, Amber uses HIE and HID which is not recognised by ANSURR
    out = open(home_dir+"/"+job_id+"/refined/anm_"+pdb_id+"_refined.pdb",'w')   
    for line in open(home_dir+"/"+job_id+"/refined/anm_"+pdb_id+"_refined.tmp",'r'):
        out.write(line.replace('HIE','HIS').replace('HID','HIS'))
    out.close()

    os.remove(home_dir+"/"+job_id+"/refined/anm_"+pdb_id+"_refined.tmp")

    # read REMARKs in combined file to generate dictionary which links model number to the name of the PDB

    model_link = {}
    for line in open(home_dir+"/"+job_id+"/refined/anm_"+pdb_id+"_refined.pdb",'r'):
        line = line.split()
        if line[0] == 'REMARK':
            model_link[str(line[2])] = line[4]
            
    with open(home_dir+"/"+job_id+"/refined/model_link.json", "w") as outfile:
        json.dump(model_link, outfile)


    # check refinement success
    if refined_pdbs > 0:
        check_quiet_print(quiet,home_dir+'/'+job_id,"sucessfully refined "+str(refined_pdbs)+" models")
    else:
        check_quiet_print(quiet,home_dir+'/'+job_id,"ERROR failed to refine any models, quitting")
        sys.exit(0)





if __name__ == "__main__":
    import sys
    mode = sys.argv[1] # why do I do this to myself?

    if mode == 'refine':
        home_dir = sys.argv[2]
        job_id = sys.argv[3]
        pdb = sys.argv[4]
        cpu = sys.argv[5]
        tleap = sys.argv[6]
        sander = sys.argv[7]
        sander_MPI = sys.argv[8]
        cpptraj = sys.argv[9]
        refine(home_dir, job_id, pdb, cpu, tleap, sander, sander_MPI, cpptraj)
    elif mode == 'combine':
        home_dir = sys.argv[2]
        job_id = sys.argv[3]
        pdb_id = sys.argv[4]
        combine(home_dir, job_id, pdb_id)



import sys
import os
import glob
import shutil
import json
import numpy as np
import re

from importlib import resources
from pdbtools import pdb_mkensemble
from nerrds.functions import check_quiet_print

with resources.path("nerrds.scripts.BME","BME.py") as f:
    BME_path = os.path.dirname(f)
sys.path.insert(1, BME_path)
import BME

natsort = lambda s: [int(t) if t.isdigit() else t.lower() for t in re.split('(\\d+)', s)] # used throughout to loop through files in natural order

def reweight_ensemble(home_dir,job_id,pdb_id,shifts_id,quiet=False):

    check_quiet_print(quiet,home_dir+'/'+job_id,"determining ensemble using BME")

    bme_dir = home_dir+"/"+job_id+"/bme/"
    os.makedirs(bme_dir,exist_ok=True)
    os.chdir(bme_dir)

    # collate ansurr output into data files for BME
    data = glob.glob(home_dir+"/"+job_id+"/ansurr/anm_"+pdb_id+"_refined_"+shifts_id+"/ANSURR_output/out/*.out")

    calc = []
    exp = []
    labels = []
    n=0
    for i in sorted(data,key=natsort):
        calc_tmp = []
        for line in open(i,'r'):
            if 'nan' not in line:
                line = line.split()
                if n==0:
                    exp.append([float(line[2]),0.05])
                    labels.append(line[0])   # do we need labels? probably yes because includes residue IDs actually used 
                    calc_tmp.append(float(line[3]))
                else:
                    calc_tmp.append(float(line[3]))
                    
        calc.append(calc_tmp)        
        n=1

    calc = np.array(calc)
    exp = np.array(exp)  

    out = open(bme_dir+"exp.txt",'w')
    # BME wants to know what kind of the data is being used so it knows how to average the data. JCOUPLINGS is just a simple mean - fine for comparing per residue flexibility
    out.write("# DATA=JCOUPLINGS\n")  
    i=0
    for e in exp:
        i+=1
        out.write(str(i)+' '+str(e[0])+' '+str(e[1])+'\n')
    out.close()

    j=0
    out = open(bme_dir+"calc.txt",'w')
    for e in calc:
        j+=1
        out.write(str(j)+' ')
        for i in e:
            out.write(str(i)+' ')
        out.write('\n')
    out.close()

    # define input file names
    exp_file = bme_dir+"exp.txt"
    calc_file = bme_dir+"calc.txt"

    # initialize. A name must be specified 
    rew = BME.Reweight("reweight_ANM")

    # load the experimental and calculated datasets
    rew.load(exp_file,calc_file)

    # 5-fold cross validation for 20 theta values
    thetas = np.geomspace(1,10000,20) # will this hold for everything?
    opt_theta = rew.theta_scan(thetas=thetas,nfold=5)

    # initialize. A name must be specified 
    rew = BME.Reweight("reweight_ANM_opt")

    # load the experimental and calculated datasets
    rew.load(exp_file,calc_file)

    # fit the data 
    chi2_before, chi2_after, phi = rew.fit(theta=opt_theta)

    # get weights
    weights = rew.get_weights()

    # copy chosen, named accoring to weight, and combined into a single file
    ensemble_dir = home_dir+"/"+job_id+"/ensemble/"
    os.makedirs(ensemble_dir,exist_ok=True)

    selected_models = [] # used later for plotting PCA
    selected_weights = [] # used later for plotting PCA

    with open(home_dir+"/"+job_id+"/refined/model_link.json") as json_file:
        model_link = json.load(json_file)

    mkensemble_string = []
    chosen_ensemble = {}
    for i in enumerate(weights):
        if i[1] > 0.01:   # have a cut off - need to have at least 1% weight
            #print(i[0]+1,round(i[1],2),model_link[str(i[0]+1)])
            chosen_ensemble[str(i[0]+1)] = {"weight":str(round(i[1],2)),"pdb":model_link[str(i[0]+1)]}
            shutil.copyfile(home_dir+"/"+job_id+"/refined/"+model_link[str(i[0]+1)].split('_refined.pdb')[0]+"/"+model_link[str(i[0]+1)],ensemble_dir+os.path.splitext(model_link[str(i[0]+1)])[0]+'_'+str(int(100*round(i[1],2)))+'.pdb')
            mkensemble_string.append(ensemble_dir+os.path.splitext(model_link[str(i[0]+1)])[0]+'_'+str(int(100*round(i[1],2)))+'.pdb')
            selected_models.append(i[0]+1)
            selected_weights.append(round(i[1],2))
            
    new_pdb = pdb_mkensemble.run(mkensemble_string)

    out = open(ensemble_dir+pdb_id+'_ensemble.pdb','w')
    for line in enumerate(new_pdb):
        out.write(line[1])
    out.close()

    if len(chosen_ensemble) > 0:
        check_quiet_print(quiet,home_dir+'/'+job_id,"determined an ensemble consisting of "+str(len(chosen_ensemble))+" models with the following weights:")
        max_pdb_name = []
        for i in chosen_ensemble:
            for pdb in chosen_ensemble[i]:
                max_pdb_name.append(int(len(pdb)))

        max_pdb_name ='{:>'+str(max(max_pdb_name))+'}'
        max_model_name = '{:>'+str(max([len(i) for i in chosen_ensemble]))+'}'

        for model in chosen_ensemble:
            check_quiet_print(quiet,home_dir+'/'+job_id," -> model "+model+" "+chosen_ensemble[model]["pdb"]+" "+chosen_ensemble[model]["weight"])
        check_quiet_print(quiet,home_dir+'/'+job_id,"the ensemble has been saved here: "+ensemble_dir+pdb_id+'_ensemble.pdb')
    else:
        check_quiet_print(quiet,home_dir+'/'+job_id,"BME failed to determine an ensemble, quitting")
        sys.exit(0)
        

if __name__ == "__main__":
    import sys,os,glob
    home_dir = sys.argv[1]
    job_id = sys.argv[2]
    pdb_id = sys.argv[3]
    shifts_id = sys.argv[4]
    reweight_ensemble(home_dir,job_id,pdb_id,shifts_id)
    
    # clean up after dag job
    delete_dag = glob.glob(home_dir+"/"+job_id+".dag*")
    for f in delete_dag:
        os.remove(f)
    delete_scripts = glob.glob(home_dir+"/"+job_id+"/*.py")
    for f in delete_scripts:
        os.remove(f)
    delete_jobs = glob.glob(home_dir+"/"+job_id+"*.sh")
    for f in delete_jobs:
        os.remove(f)
    delete_jobs = glob.glob(home_dir+"/"+job_id+"*.sub")
    for f in delete_jobs:
        os.remove(f)


from __future__ import print_function

import sys,os
import time
import numpy as np
import pandas as pd
#import matplotlib.pyplot as plt
from scipy.optimize import minimize #version >0.13.0 for dogleg optimizer
from scipy.stats import linregress
from scipy.special import logsumexp
from sklearn.linear_model import LinearRegression
#import warnings
import BME_tools as bt
#import csv

known_methods = ["BME","BER","CHI2_L2","CHI1_L1"]

def progress(count, total, suffix=''):
    total -= 1
    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    sys.stdout.write('[%s] %s%s ...%s\r' % (bar, percents, '%', suffix))
    sys.stdout.flush()
    
# reweight class. 
class Reweight:

    # initialize
    def __init__(self,name,w0=[]):

        self.name = name

        if(len(w0)!=0):
            self.w0 = w0/np.sum(w0)
        else:
            self.w0= []
        
        self.w_opt = []
        self.lambdas = []
        
        self.log = open("%s.log" % name, "w")
        
        self.labels = []
        self.experiment =  []
        self.calculated =  []
        self.standardized = False
        
    def read_file(self,exp_file,calc_file,averaging="auto",fit='no',use_samples=[],use_data=[]):

        # read file
        log = ""
        label,exp,calc,log,averaging = bt.parse(exp_file,calc_file,averaging=averaging)
        self.log.write(log)
        
        # remove datapoints if use_samples or use_data is not empty
        label,exp, calc, log = bt.subsample(label,exp,calc,use_samples,use_data)
        self.log.write(log)

        if(len(self.w0)==0):
            self.w0 = np.ones(calc.shape[0])/calc.shape[0]
            self.log.write("Initialized uniform weights %d\n" % calc.shape[0])

        # fit/scale
        calc_avg,log = bt.fit_and_scale(exp,calc,self.w0,fit=fit)
        self.log.write(log)

        # do sanity checks
        log  = bt.check_data(label,exp,calc,self.w0)
        self.log.write(log)
        

        return label,exp,calc

    def load(self,exp_file,calc_file,averaging="auto",fit='no',use_samples=[],use_data=[],weight=1):
        
        label,exp, calc = self.read_file(exp_file,calc_file,averaging=averaging,fit=fit,use_samples=use_samples,use_data=use_data)

        if(len(self.experiment)==0):
            self.experiment = exp
            self.calculated = calc
            self.labels = label
            self.weights = np.ones(exp.shape[0])*weight
        else:
            self.experiment = np.vstack([self.experiment,exp])
            self.calculated = np.hstack([self.calculated,calc])
            self.labels = np.hstack([self.labels,label])
            self.weights = np.hstack([self.weights,np.ones(exp.shape[0])*weight])
        
        # note to self: implement weight

    # add data from external array
    def load_array(self,label,exp,calc,weight=1):
        
        if(len(self.experiment)==0):
            self.experiment = exp
            self.calculated = calc
            self.labels = label
            if(len(self.w0)==0):
                self.w0 = np.ones(calc.shape[0])/calc.shape[0]
            self.weights = np.ones(exp.shape[0])*weight
        else:
            self.experiment = np.vstack([self.experiment,exp])
            self.calculated = np.hstack([self.calculated,calc])
            self.labels = np.hstack([self.labels,label])
            self.weights = np.hstack([self.weights,np.ones(exp.shape[0])*weight])
            
    def predict(self,exp_file,calc_file,outfile=None,averaging="auto",fit='no',use_samples=[],use_data=[]):

        label,exp,calc,log,averaging = bt.parse(exp_file,calc_file,averaging=averaging)
        
        self.log.write(log)
        
        # remove datapoints if use_samples or use_data is not empty
        label,exp, calc, log = bt.subsample(label,exp,calc,use_samples,use_data)
        self.log.write(log)

        # do sanity checks
        stats,log  = bt.calc_stats(label,exp,calc,self.w0,self.w_opt,averaging=averaging,fit=fit,outfile=outfile)
        self.log.write(log)
        
        return stats
    
    def predict_array(self,label,exp,calc,outfile=None,averaging="linear",fit="no"):
        stats, log = bt.calc_stats(label,exp,calc,self.w0,self.w_opt,averaging=averaging,outfile=outfile,fit=fit)
        return stats


        # do sanity checks
        #log  = bt.check_data(label,exp,calc,self.w_opt)
        #self.log.write(log)

    def get_lambdas(self):
        return self.lambdas
    
    def get_iterations(self):
        return self.niter

    def get_nsamples(self):
        return self.calculated.shape[0]

    def get_ndata(self):
        return self.experiment.shape[0]

    def get_labels(self):
        return self.labels
    
    def get_experiment(self):
        return np.copy(self.experiment)
    
    def get_calculated(self):
        return np.copy(self.calculated)

    def get_name(self):
        return self.name
    
    def get_weights(self):
        return np.copy(self.w_opt)
    
    def get_w0(self):
        return np.copy(self.w0)

    def set_lambdas(self,lambda0):
        if(len(self.lambdas)==0):
            self.lambdas = lambda0
        else:
            print("# Overriding lambdas is not possible")
            sys.exit(1)


    # optimize
    def fit(self,theta,lambdas_init=True,method="BME"):

        if(self.standardized==False):
            log,v1,v2 = bt.standardize(self.experiment,self.calculated,self.w0,normalize="zscore")
            self.standardized=True

        assert method in known_methods, "method %s not in known methods:" % (method,known_methods)
        
        def maxent(lambdas):
            # weights
            arg = -np.sum(lambdas[np.newaxis,:]*self.calculated,axis=1)-tmax+np.log(self.w0)
            
            logz = logsumexp(arg)
            ww = np.exp(arg-logz)
            
            avg = np.sum(ww[:,np.newaxis]*self.calculated, axis=0)
            
            # gaussian integral
            eps2 = 0.5*np.sum((lambdas*lambdas)*theta_sigma2) 
            
            # experimental value 
            sum1 = np.dot(lambdas,self.experiment[:,0])
            #fun = sum1 + eps2 + np.log(zz)
            fun = sum1 + eps2 + logz
            
            # gradient
            #jac = self.experiment["avg"].values + lambdas*err - avg
            jac = self.experiment[:,0] + lambdas*theta_sigma2 - avg
            

            # divide by theta to avoid numerical problems
            return  fun/theta,jac/theta

    
        def maxent_hess(lambdas):
            arg = -np.sum(lambdas[np.newaxis,:]*self.calculated,axis=1) -tmax
            #arg -= tmax
            
            #########
            ww = (self.w0*np.exp(arg))
            zz = np.sum(ww)
            assert np.isfinite(zz), "# Error. sum of weights is infinite. Use higher theta"
            ww /= zz

            q_w = np.dot(ww,self.calculated)
            hess = np.einsum('k, ki, kj->ij',ww,self.calculated,self.calculated) - np.outer(q_w,q_w) + np.diag(theta_sigma2)

            return  hess/theta

        def func_ber_gauss(w):

            bcalc = np.sum(w[:,np.newaxis]*self.calculated,axis=0)
            diff = bcalc-self.experiment[:,0]
            #print(diff)
            #ii = np.where(((diff<0) & (self.experiment[:,2]<0)) | ((diff>0) & (self.experiment[:,2]>0)) )[0#]
 #           ff = [1 if (self.experiment[j,2]==0 or j in ii) else 0 for j in range(self.experiment.shape[0])]
            #diff *= ff
            
            chi2_half =  0.5*np.sum(((diff/self.experiment[:,1])**2))

            
            idxs = np.where(w>1.0E-50)
            log_div = np.zeros(w.shape[0])
            log_div[idxs] = np.log(w[idxs]/self.w0[idxs])
            srel = theta*np.sum(w*log_div)

            jac = np.sum(diff*self.calculated,axis=1) + theta*(1.+log_div)
            return chi2_half+srel#, jac
    

        
        def func_chi2_L2(w):

            bcalc = np.sum(w[:,np.newaxis]*self.calculated,axis=0)
            diff = (bcalc-self.experiment[:,0])/self.experiment[:,1]
            
            ii = np.where(((diff<0) & (self.experiment[:,2]<0)) | ((diff>0) & (self.experiment[:,2]>0)) )[0]
            ff = [1 if (self.experiment[j,2]==0 or j in ii) else 0 for j in range(self.experiment.shape[0])]

            diff *= ff
            
            chi2_half =  0.5*np.sum(diff**2)

            jac = np.sum(diff*self.calculated,axis=1)
            #idxs = np.where(w>1.0E-50)
                                                                                                                                              #srel = theta*np.sum(w[idxs]*np.log(w[idxs]/self.w0[idxs]))
                                                                                                                                              #jac = 
            return chi2_half
        
        def func_chi2_L1(w):

            bcalc = np.sum(w[:,np.newaxis]*self.calculated,axis=0)
            diff = (bcalc-self.experiment[:,0])/self.experiment[:,1]
            
            ii = np.where(((diff<0) & (self.experiment[:,2]<0)) | ((diff>0) & (self.experiment[:,2]>0)) )[0]
            ff = [1 if (self.experiment[j,2]==0 or j in ii) else 0 for j in range(self.experiment.shape[0])]

            diff *= ff
            
            chi2_half =  0.5*np.sum(diff**2)

            jac = np.sum(diff*self.calculated,axis=1)
            #idxs = np.where(w>1.0E-50)
                                                                                                                                              #srel = theta*np.sum(w[idxs]*np.log(w[idxs]/self.w0[idxs]))
                                                                                                                                              #jac = 
            return chi2_half,jac

        if(lambdas_init==True):
            lambdas=np.zeros(self.experiment.shape[0],dtype=np.longdouble)
            self.log.write("Lagrange multipliers initialized from zero\n")
        else:
            assert(len(self.lambdas)==self.experiment.shape[0])
            lambdas = np.copy(self.lambdas)
            #np.array(lambdas_init)
            self.log.write("Warm start\n")
        #print(lambdas)
            
        bounds = []
        for j in range(self.experiment.shape[0]):
            if(self.experiment[j,2]==0):
                bounds.append([None,None])
            elif(self.experiment[j,2]==-1):
                bounds.append([None,0.0])
            else:
                bounds.append([0.0,None])

        if(method=="BME"):
            opt={'maxiter':50000,'disp':False}
            
            tmax = np.log((sys.float_info.max)/5.)

            theta_sigma2 = theta*self.weights*self.experiment[:,1]**2

            chi2_before  = bt.calc_chi(self.experiment,self.calculated,self.w0)
            self.log.write("Optimizing %d data and %d samples. Theta=%f \n" % (self.experiment.shape[0],self.calculated.shape[0],theta))
            self.log.write("CHI2 before optimization: %8.4f \n" % (chi2_before))
            self.log.flush()
            mini_method = "L-BFGS-B"
            start_time = time.time()
            #if(all(self.experiment[:,2]==0)):
            #    mini_method="trust-constr"
            #    result = minimize(maxent,lambdas,\
            #                               options=opt,method=mini_method,\
            #                               jac=True,hess=maxent_hess)
            
            result = minimize(maxent,lambdas,\
                              options=opt,method=mini_method,\
                              jac=True,bounds=bounds)
            self.log.write("Execution time: %.2f seconds\n" % (time.time() - start_time))
            
            if(result.success):
                self.log.write("Minimization using %s successful (iterations:%d)\n" % (mini_method,result.nit))
                arg = -np.sum(result.x[np.newaxis,:]*self.calculated,axis=1) -tmax
                w_opt = self.w0*np.exp(arg)
                w_opt /= np.sum(w_opt)
                self.lambdas = np.copy(result.x)
                self.w_opt = np.copy(w_opt)
                self.niter = result.nit
                chi2_after  = bt.calc_chi(self.experiment,self.calculated,w_opt)
                phi = np.exp(-bt.srel(self.w0,w_opt))
                
                self.log.write("CHI2 after optimization: %8.4f \n" % (chi2_after))
                self.log.write("Fraction of effective frames: %8.4f \n" % (phi))
                self.log.flush()
                return chi2_before,chi2_after,phi
            
            else:
                self.log.write("Minimization using %s failed\n" % (mini_method))
                self.log.write("Message: %s\n" % (result.message))
                self.niter = -1
                self.log.flush()
                return np.NaN, np.NaN, np.NaN
            
            

        # please check 
        if(method=="BER"):
            
            opt={'maxiter':2000,'disp': True,'ftol':1.0e-20}
            cons = {'type': 'eq', 'fun':lambda x: np.sum(x)-1.0}
            bounds = [(0.,None)]*len(self.w0)
            mini_method = "SLSQP"
            chi2_before  = bt.calc_chi(self.experiment,self.calculated,self.w0)
            self.log.write("CHI2 before optimization: %8.4f \n" % (chi2_before))
            self.log.flush()
                                                                                                                                            
            w0 = np.copy(self.w0)
            start_time = time.time()
            #print(func_ber_gauss(w0))
            #result = minimize(func_ber_gauss,w0,constraints=cons,options=opt,method=mini_method,bounds=bounds,jac=True)
            result = minimize(func_ber_gauss,w0,constraints=cons,options=opt,method=mini_method,bounds=bounds,jac=False)
            self.log.write("Execution time: %.2f seconds\n" % (time.time() - start_time))
            if(result.success):
                self.log.write("Minimization using %s successful (iterations:%d)\n" % (mini_method,result.nit))
                w_opt = np.copy(result.x)
                self.w_opt = w_opt
                chi2_after  = bt.calc_chi(self.experiment,self.calculated,w_opt)
                phi = np.exp(-bt.srel(self.w0,w_opt))
                self.log.write("CHI2 after optimization: %8.4f \n" % (chi2_after))
                self.log.write("Fraction of effective frames: %8.4f \n" % (phi))
                self.log.flush()
                return chi2_before,chi2_after,phi
            
            else:
                self.log.write("Minimization using %s failed\n" % (mini_method))
                self.log.write("Message: %s\n" % (result.message))
                self.log.flush()
                return np.NaN, np.NaN, np.NaN
   
        # please check 
        if(method=="CHI2_L2"):
            
            opt={'maxiter':2000,'disp': True,'ftol':1.0e-20}
            cons = {'type': 'eq', 'fun':lambda x: np.sum(x)-1.0}
            bounds = [(0.,None)]*len(self.w0)
            meth = "SLSQP"
            chi2_before  = bt.calc_chi(self.experiment,self.calculated,self.w0)
            self.log.write("CHI2 before optimization: %8.4f \n" % (chi2_before))
            start_time = time.time()
            result = minimize(func_chi2_L2,self.w0,constraints=cons,options=opt,method=meth,jac=True,bounds=bounds)

            w_opt = np.copy(result.x)
            self.w_opt = w_opt
            chi2_after  = bt.calc_chi(self.experiment,self.calculated,w_opt)
            phi = np.exp(-bt.srel(self.w0,w_opt))
            self.log.write("Execution time: %.2f seconds\n" % (time.time() - start_time))
            self.log.write("CHI2 after optimization: %8.4f \n" % (chi2_after))

        # please check 
        if(method=="CHI2_L1"):
            
            opt={'maxiter':2000,'disp': True,'ftol':1.0e-20}
            cons = {'type': 'eq', 'fun':lambda x: np.sum(x)-1.0}
            bounds = [(0.,None)]*len(self.w0)
            meth = "SLSQP"
            chi2_before  = bt.calc_chi(self.experiment,self.calculated,self.w0)
            self.log.write("CHI2 before optimization: %8.4f \n" % (chi2_before))
            start_time = time.time()
            result = minimize(func_chi2_L2,self.w0,constraints=cons,options=opt,method=meth,jac=True,bounds=bounds)

            w_opt = np.copy(result.x)
            self.w_opt = w_opt
            chi2_after  = bt.calc_chi(self.experiment,self.calculated,w_opt)
            phi = np.exp(-bt.srel(self.w0,w_opt))
            self.log.write("Execution time: %.2f seconds\n" % (time.time() - start_time))
            self.log.write("CHI2 after optimization: %8.4f \n" % (chi2_after))



    def theta_scan(self,thetas=[],train_fraction_data=0.75,nfold=5,train_fraction_samples=0.8):

        np.random.seed(42)
        tmp_dir = "theta_scan"
        os.system("mkdir -p %s" % tmp_dir)
        if(len(thetas)==0): thetas = np.geomspace(0.1,10000,10)
        #print("Performing %d-fold cross validation for %d theta values" % (nfold,len(thetas)))
        #print("Output files are written to %s" % (tmp_dir))
              
        nsamples = self.get_nsamples()
        ndata =   self.get_ndata()
        
        train_samples =  int(nsamples*train_fraction_samples)
        train_data =  int(ndata*train_fraction_data)

        results = np.zeros((len(thetas),nfold,3))
        for i in range(nfold):
            #progress(i,nfold)
            #if(i==nfold-1):
            #    print("\n")
            shuffle_samples = np.arange(nsamples)
            shuffle_data = np.arange(ndata)
            np.random.shuffle(shuffle_samples)
            np.random.shuffle(shuffle_data)
            
            train_idx_data = shuffle_data[:train_data]
            train_idx_samples = shuffle_samples[:train_samples]
            test_idx_data = shuffle_data[train_data:]
            test_idx_samples = shuffle_samples[:train_samples]  # test samples are the same as train!
        

            labels_train = [self.labels[k] for k in train_idx_data]
            labels_test = [self.labels[k] for k in test_idx_data]
            
            exp = self.get_experiment()
            calc = self.get_calculated()
            exp_train = exp[train_idx_data,:]
            calc_train = calc[:,train_idx_data]
        
            exp_test = exp[test_idx_data,:]
            calc_test = calc[:,test_idx_data]
        
            r1 = Reweight("crossval_%s_%d" % (self.name,i),w0=np.copy(self.w0[train_idx_samples]))        
            r1.load_array(labels_train,np.copy(exp_train),np.copy(calc_train[train_idx_samples,:]))
            l_init = True
        
            for j,t in enumerate(thetas):
                c1,c2,phi = r1.fit(t,lambdas_init=l_init)
                l_init = False
                fr = "crossval_%s_t%.2f_f%d" % (self.name,t,i)
                train_stats = r1.predict_array(labels_train,exp_train,calc_train[train_idx_samples,:],\
                                           outfile="%s/%s_train" % (tmp_dir,fr))
                test_stats = r1.predict_array(labels_test,exp_test,calc_test[test_idx_samples,:],
                                          outfile="%s/%s_test" % (tmp_dir,fr))
                #print("####",i,j)
                #outfiles.append(fr)
                results[j,i,0] = train_stats[3]/train_stats[0]
                results[j,i,1] = test_stats[3]/test_stats[0]
                results[j,i,2] = phi

        #fig,ax = plt.subplots(1,1,figsize=(12,8))
        #for k in range(nfold):
        #    plt.scatter(thetas,results[:,k,0],c='k',s=4)
        #    plt.scatter(thetas,results[:,k,1],c='r',s=4)
        #    plt.scatter(thetas,results[:,k,2],c='b',s=4)

        avg_test_error = np.nanmean(results[:,:,1],axis=1)
        avg_train_error = np.nanmean(results[:,:,0],axis=1)
        avg_phi = np.nanmean(results[:,:,2],axis=1)
        argmin = np.argmin(avg_test_error)
        
        #print("Optimal theta: %.2f" % thetas[argmin])
        #print("Validation error reduction %.3f" % avg_test_error[argmin])
        #print("Training error reduction %.3f" % avg_train_error[argmin])
        #print("Fraction of effective frames  %.3f" % avg_phi[argmin])

        #plt.plot(thetas,avg_train_error,"-*",c='k',label="Training error")
        #plt.plot(thetas,avg_test_error,"-o",c='r',label="Test error")
        #plt.plot(thetas,avg_phi,"-o",c='b',label="Phi")
        #plt.legend()
        #if(np.max(results[:,:,1])<1.):
        #    ax.set_ylim(0,1.1)
        #else:
        #    ax.set_ylim(0,np.min([2,1.1*np.max(results[:,:,1])]))
        #plt.scatter(thetas[argmin],avg_train_error[argmin],c='r')
        #ax.axhline(1,ls="--",color='0.4')
        #ax.set_xscale('log')
        #ax.set_xlabel("Theta")
        #plt.savefig("%s/crossval_%s.png" % (tmp_dir,self.name))

        return thetas[argmin]
        
        
    def ibme(self,theta,ftol=0.01,iterations=50,lr_weights=True,offset=True):
        
        current_weights = self.get_w0()
        w0 = self.get_w0()
        name = self.get_name()
        labels = self.get_labels()
        exp = self.get_experiment()
        calc = self.get_calculated()

        self.ibme_weights = []
        self.ibme_stats = []
        
        #print("# iterative BME")
        if(lr_weights==True):
            inv_var = 1./exp[:,1]**2
        else:
            inv_var = np.ones(len(exp))

        log = ""
        rr_old = np.NaN
        #self.log.write("acascacasacas\n")
        for it in range(iterations):
            
            calc_avg = np.sum(calc*current_weights[:,np.newaxis],axis=0)

            model = LinearRegression(fit_intercept=offset)
            model.fit(calc_avg.reshape(-1,1),exp[:,0],inv_var)

            #alpha, beta, r_value, p_value, std_err = linregress(calc_avg,exp[:,0])
            alpha = model.coef_[0] #Scale factor
            beta = model.intercept_
            calc = alpha*calc+beta
            
            r1 = Reweight("%s_ibme_%d" % (name,it),w0=np.copy(w0))
            r1.load_array(labels,np.copy(exp),np.copy(calc))
            rr= r1.fit(theta=theta)
            if(it==0): chi2_0 = rr[0]
            if(it==0): calc_0 = np.copy(calc)
                
            current_weights = np.copy(r1.get_weights())

            diff = abs(rr_old-rr[1])
            line = "Iteration:%3d scale: %7.4f offset: %7.4f chi2: %7.4f diff: %7.4e\n" % (it,alpha,beta,rr[1],diff)
            rr_old = rr[1]
            #print(line,end="")
            log += line

            self.ibme_weights.append(current_weights)
            self.ibme_stats.append(rr)
            
            if(diff<ftol):
                line = "Iterative procedure converged below tolerance %.2e after %d iterations\n" % (diff,it)
                print(line,end="")
                log += line
                break
            
        self.log.write(log+ "\n")
        #self.log.close()
        self.log.flush()
        n1 = "%s_%d.calc.dat" % (self.name,it)
        n2 = "%s_%d.weights.dat" % (self.name,it)
        
        df = pd.DataFrame(calc)
        df.to_csv(n1,sep=" ",header=False,float_format="%8.4e")

        df = pd.DataFrame(current_weights)
        df.to_csv(n2,sep=" ",header=False,float_format="%8.4e")
        
        #print("Done. Initial chi2: %8.4f Final chi2:%8.4f" % (chi2_0,rr[1]))
        #print("Done. Writing output files %s %s" % (n1,n2))
        phi = np.exp(-bt.srel(w0,current_weights))
        self.w_opt = current_weights
        return chi2_0,rr[1],phi,calc_0,calc

    def get_ibme_weights(self):

        try:
            return self.ibme_weights
        except:
            print("# iBME weights not available. Call iBME first")
            sys.exit(1)

    def get_ibme_stats(self):

        try:
            return self.ibme_stats
        except:
            print("# iBME stats not available. Call iBME first")
            sys.exit(1)

        
    def iterative_theta_scan(self,thetas=[],ftol=0.01,iterations=50,lr_weights=True,offset=True,
                            train_fraction_data=0.75,nfold=5,train_fraction_samples=0.8):

        np.random.seed(42)
        if(len(thetas)==0): thetas = np.geomspace(0.1,10000,10)
        #print("Performing %d-fold cross validation for %d theta values" % (nfold,len(thetas)))
        nsamples = self.get_nsamples()
        ndata =   self.get_ndata()
        
        train_samples =  int(nsamples*train_fraction_samples)
        train_data =  int(ndata*train_fraction_data)

        results = np.zeros((len(thetas),nfold,3))
        for i in range(nfold):
            #progress(i,nfold)
            #if(i==nfold-1):
            #    print("\n")
            shuffle_samples = np.arange(nsamples)
            shuffle_data = np.arange(ndata)
            np.random.shuffle(shuffle_samples)
            np.random.shuffle(shuffle_data)
            
            train_idx_data = shuffle_data[:train_data]
            train_idx_samples = shuffle_samples[:train_samples]
            test_idx_data = shuffle_data[train_data:]
            test_idx_samples = shuffle_samples[:train_samples]  # test samples are the same as train!
        

            #labels_train = [self.labels[k] for k in train_idx_data]
            #labels_test = [self.labels[k] for k in test_idx_data]
            
            exp = self.get_experiment()
            calc = self.get_calculated()
            exp_train = exp[train_idx_data,:]
            calc_train = calc[:,train_idx_data]
        
            exp_test = exp[test_idx_data,:]
            calc_test = calc[:,test_idx_data]

            r1 = Reweight("crossval_%s_%d" % (self.name,i),w0=np.copy(self.w0[train_idx_samples]))       
            r1.load_array(train_idx_data,np.copy(exp_train),np.copy(calc_train[train_idx_samples,:]))
        
            for j,t in enumerate(thetas):
                c1,c2,phi,calc_0,calc_1 = r1.ibme(t,ftol=ftol,iterations=iterations,
                                    lr_weights=lr_weights,offset=offset)
                
                #calc = a*calc+b
                #calc_train = calc[:,train_idx_data]
                #calc_test = calc[:,test_idx_data]

                l_init = False
                fr = "crossval_%s_t%.2f_f%d" % (self.name,t,i)
                chi_train_0 = bt.calc_chi(exp[train_idx_data],calc_0, r1.w0)
                chi_train_1 = bt.calc_chi(exp[train_idx_data],calc_1, r1.w_opt)
                #Because the alpha and beta parameters for the scaled observables are not kept in ibme,
                #we fit the final calc arrays to obtain them.                
                model = LinearRegression(fit_intercept=offset)
                model.fit(calc_train[train_idx_samples,0].reshape(-1,1),calc_0[:,0])
                # Initial values
                alpha = model.coef_[0] #Scale factor
                beta = model.intercept_
                calc_test_0 = alpha*calc_test+beta
                # Optimized values
                model.fit(calc_train[train_idx_samples,0].reshape(-1,1),calc_1[:,0])
                alpha = model.coef_[0] #Scale factor
                beta = model.intercept_
                calc_test_1 = alpha*calc_test+beta

                chi_test_0 = bt.calc_chi(exp[test_idx_data],calc_test_0[test_idx_samples,:], r1.w0)
                chi_test_1 = bt.calc_chi(exp[test_idx_data],calc_test_1[test_idx_samples,:], r1.w_opt)
                #print("####",i,j)
                #outfiles.append(fr)
                results[j,i,0] = chi_train_1/chi_train_0
                results[j,i,1] = chi_test_1/chi_test_0
                results[j,i,2] = phi

        #fig,ax = plt.subplots(1,1,figsize=(12,8))
        #for k in range(nfold):
        #    plt.scatter(thetas,results[:,k,0],c='k',s=4)
        #    plt.scatter(thetas,results[:,k,1],c='r',s=4)
        #    plt.scatter(thetas,results[:,k,2],c='b',s=4)

        avg_test_error = np.nanmean(results[:,:,1],axis=1)
        avg_train_error = np.nanmean(results[:,:,0],axis=1)
        avg_phi = np.nanmean(results[:,:,2],axis=1)
        argmin = np.argmin(avg_test_error)
        
        #print("Optimal theta: %.2f" % thetas[argmin])
        #print("Validation error reduction %.3f" % avg_test_error[argmin])
        #print("Training error reduction %.3f" % avg_train_error[argmin])
        #print("Fraction of effective frames  %.3f" % avg_phi[argmin])

        #plt.plot(thetas,avg_train_error,"-*",c='k',label="Training error")
        #plt.plot(thetas,avg_test_error,"-o",c='r',label="Test error")
        #plt.plot(thetas,avg_phi,"-o",c='b',label="Phi")
        #plt.legend()
        #if(np.max(results[:,:,1])<1.):
        #    ax.set_ylim(0,1.1)
        #else:
        #    ax.set_ylim(0,np.min([2,1.1*np.max(results[:,:,1])]))
        #plt.scatter(thetas[argmin],avg_train_error[argmin],c='r')
        #ax.axhline(1,ls="--",color='0.4')
        #ax.set_xscale('log')
        #ax.set_xlabel("Theta")
        #plt.savefig("crossval_%s.png" % self.name)

        return thetas[argmin]
        




import sys
import os
import argparse
import shutil
import glob
import re
import subprocess

from importlib import resources
from datetime import datetime

from nerrds import gen_conf
from nerrds import refine
from nerrds import run_ansurr
from nerrds import reweight_bme
from nerrds.functions import check_quiet_print

nerrds_version = "0.0.105"

def main():

    #------------------------------------------------------ check if AMBERtools is available and quit if not ------------------------------------------------
     
    tleap = shutil.which("tleap")
    sander = shutil.which("sander")
    sander_MPI = shutil.which("sander.MPI")
    cpptraj = shutil.which("cpptraj")

    if sander is None and sander_MPI is None:
        print("ERROR could not find sander or sander.MPI (needed to refine models), quitting")
        sys.exit(0)

    elif tleap == 0:
        print("ERROR could not find tleap (needed to refine models), quitting")
        sys.exit(0)

    elif cpptraj == 0:
        print("ERROR could not find cpptraj (needed to refine models), quitting")
        sys.exit(0)

    #-------------------------------------------------------------- args ----------------------------------------------------------------------------

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--pdb", type=str, help="input PDB file",required=True)
    parser.add_argument("-s", "--shifts", type=str, help="input shifts file in NMR-STAR v3 or NEF format",required=True)

    parser.add_argument("-c", "--conf", help="number of conformers to generate per model in input ensemble (default 50)")
    parser.add_argument("-r", "--avrmsd", help="average RMSD of conformers generated per model in input ensemble (default 2A)")
    parser.add_argument("-n", "--ncpu", help="number of CPUs used to refine models in AMBER (default 1)")
    parser.add_argument("-e", "--env", help="path to virtual environment on NMRbox, needed to run refinement in parallel via HTCondor")
    parser.add_argument("-q", "--quiet", help="suppress output to the terminal", action="store_true")
    parser.add_argument("-v", "--version", help="NERRDS | NMR Ensemble Refinement using RigiDity anf Shifts v"+nerrds_version)

    args = parser.parse_args()

    #-------------------------------------------------------------- parameters ----------------------------------------------------------------------------

    # number of conformers to generate per model in input ensemble
    if args.conf:
        conformers = args.conf
    else:
        conformers = 50 

    # average RMSD of confomers generated
    if args.avrmsd:
        average_rmsd = args.avrmsd
    else:
        average_rmsd = 2    

    # number of CPUs used to refine models in AMBER
    if args.ncpu:
        if sander_MPI is None:
            if sander is not None:
                print("WARNING cannot find sander.MPI to run refinement with multiple cpus, continuing with sander and ncpu=1")
                cpu = 1
        else:
            cpu = int(args.ncpu)
            if cpu == 1 and sander is None:
                print("ERROR cannot find sander to run refinement with 1 cpu, quitting")
                sys.exit(0)
    else:
        if sander is None:
            print("ERROR cannot find sander to run refinement with 1 cpu, quitting")
            sys.exit(0)
        else:
            cpu = 1

    #-------------------------------------------------------------- variables ----------------------------------------------------------------------------

    home_dir = os.getcwd()

    # shifts and PDB ensemble required as input
    path_to_pdb = home_dir+"/"+args.pdb
    path_to_shifts = home_dir+"/"+args.shifts

    # useful for later when running ansurr
    pdb_id = os.path.splitext(os.path.basename(path_to_pdb))[0]    
    shifts_id = os.path.splitext(os.path.basename(path_to_shifts))[0]

    # unique jobID based on time - could append pdb_id?
    job_id = datetime.now().strftime("%Y"+"%m"+"%d"+"_"+"%H"+"%M"+"%S"+"%f"+"_"+pdb_id+"_"+shifts_id)

    natsort = lambda s: [int(t) if t.isdigit() else t.lower() for t in re.split('(\\d+)', s)] # used throughout to loop through files in natural order

    now = datetime.now()


    os.makedirs(job_id,exist_ok=True)

    log_file = open(home_dir+'/'+job_id+'/NERRDS.log','w')
    log_file.write('\n')
    log_file.close()
    check_quiet_print(args.quiet,home_dir+'/'+job_id,"\nNERRDS v"+nerrds_version+" "+now.strftime("%d/%m/%Y %H:%M:%S"))
    check_quiet_print(args.quiet,home_dir+'/'+job_id," -> pdb: "+path_to_pdb)
    check_quiet_print(args.quiet,home_dir+'/'+job_id," -> shifts: "+path_to_shifts)
    
    if args.env is not None: # check if job to be run on HT condor

        gen_conf.gen_conf(path_to_pdb, home_dir, job_id, pdb_id, conformers, average_rmsd)

        os.makedirs("logs",exist_ok=True)

        with resources.path("nerrds", "refine.py") as f:
        	refine_script = f

        with resources.path("nerrds", "run_ansurr.py") as f:
            ansurr_script = f

        with resources.path("nerrds", "reweight_bme.py") as f:
            bme_script = f

        shutil.copy(refine_script, home_dir+'/'+job_id+"/") #copy local version as permission being denied?
        shutil.copy(ansurr_script, home_dir+'/'+job_id+"/") #copy local version as permission being denied?
        shutil.copy(bme_script, home_dir+'/'+job_id+"/") #copy local version as permission being denied?

        refine_script = home_dir+'/'+job_id+"/refine.py"
        ansurr_script = home_dir+'/'+job_id+"/run_ansurr.py"
        bme_script = home_dir+'/'+job_id+"/reweight_bme.py"

        out = open("refine.tmp","w")
        out.write("#!"+str(os.path.abspath(args.env))+'/bin/python3\n') # append specific version of python needed
        for line in open(refine_script,"r"):
            out.write(line)
        out.close()
        shutil.move("refine.tmp",refine_script) 
        os.chmod(refine_script , 0o777)

        out = open("run_ansurr.tmp","w")
        out.write("#!"+str(os.path.abspath(args.env))+'/bin/python3\n') # append specific version of python needed
        for line in open(ansurr_script,"r"):
            out.write(line)
        out.close()
        shutil.move("run_ansurr.tmp",ansurr_script) 
        os.chmod(ansurr_script , 0o777)

        out = open("reweight_bme.tmp","w")
        out.write("#!"+str(os.path.abspath(args.env))+'/bin/python3\n') # append specific version of python needed
        for line in open(bme_script,"r"):
            out.write(line)
        out.close()
        shutil.move("reweight_bme.tmp",bme_script) 
        os.chmod(bme_script , 0o777)
        
        condor_refine = open(job_id+"_run_refine.sub","w")  ###  scripts dir will need to read from package install
        condor_refine.write("executable = "+job_id+"_run_refine.sh\nrequest_memory = 4G\nrequest_cpus = "+str(cpu)+"\nrequest_gpus = 0\noutput = logs/$(JobName).stdout\nerror = logs/$(JobName).stderr\nlog = logs/refine.log\n")
        condor_refine.write("transfer_executable = FALSE\nshould_transfer_files = NO\nJobName = run_refine\nrequirements = (Target.Production == True)\n+Production = True\n")
        for pdb in sorted(glob.glob(home_dir+'/'+job_id+"/anm/anm_"+pdb_id+"_*.pdb"),key=natsort):
            p = os.path.splitext(os.path.basename(pdb))[0]
            condor_refine.write("JobName = "+p+'\n')
            condor_refine.write("arguments = refine "+home_dir+" "+job_id+" "+pdb+" "+str(cpu)+" "+str(tleap)+" "+str(sander)+" "+str(sander_MPI)+" "+str(cpptraj)+" "+'\n')
            condor_refine.write("queue"+'\n\n')
        condor_refine.close()

        condor_combine = open(job_id+"_run_combine.sub","w")  ###  scripts dir will need to read from package install
        condor_combine.write("executable = "+job_id+"_run_refine.sh\nrequest_memory = 4G\nrequest_cpus = 1\nrequest_gpus = 0\noutput = logs/$(JobName).stdout\nerror = logs/$(JobName).stderr\nlog = logs/combine.log\n")
        condor_combine.write("transfer_executable = FALSE\nshould_transfer_files = NO\nJobName = run_combine\nrequirements = (Target.Production == True)\n+Production = True\n")
        condor_combine.write("JobName = combine\n")
        condor_combine.write("arguments = combine "+home_dir+" "+job_id+" "+pdb_id+'\n')
        condor_combine.write("queue"+'\n\n')           
        condor_combine.close()

        condor_ansurr = open(job_id+"_run_ansurr.sub","w")  ###  scripts dir will need to read from package install
        condor_ansurr.write("executable = "+job_id+"_run_ansurr.sh\nrequest_memory = 4G\nrequest_cpus = 1\nrequest_gpus = 0\noutput = logs/$(JobName).stdout\nerror = logs/$(JobName).stderr\nlog = logs/ansurr.log\n")
        condor_ansurr.write("transfer_executable = FALSE\nshould_transfer_files = NO\nJobName = run_combine\nrequirements = (Target.Production == True)\n+Production = True\n")
        condor_ansurr.write("JobName = ansurr\n")
        condor_ansurr.write("arguments = "+home_dir+" "+job_id+" "+pdb_id+" "+path_to_shifts+'\n')
        condor_ansurr.write("queue"+'\n\n')
        condor_ansurr.close()

        condor_bme = open(job_id+"_run_bme.sub","w")  ###  scripts dir will need to read from package install
        condor_bme.write("executable = "+job_id+"_run_bme.sh\nrequest_memory = 4G\nrequest_cpus = 1\nrequest_gpus = 0\noutput = logs/$(JobName).stdout\nerror = logs/$(JobName).stderr\nlog = logs/bme.log\n")
        condor_bme.write("transfer_executable = FALSE\nshould_transfer_files = NO\nJobName = run_bme\nrequirements = (Target.Production == True)\n+Production = True\n")
        condor_bme.write("JobName = bme\n")
        condor_bme.write("arguments = "+home_dir+" "+job_id+" "+pdb_id+" "+shifts_id+'\n')
        condor_bme.write("queue"+'\n\n')
        condor_bme.close()

        bash_refine = open(job_id+"_run_refine.sh","w")
        bash_refine.write("#!/bin/bash\n")
        bash_refine.write("source "+str(os.path.abspath(args.env))+"/bin/activate\n")  
        bash_refine.write("python3 "+str(refine_script)+" $1 $2 $3 $4 $5 $6 $7 $8 $9"+'\n')
        bash_refine.close()
        os.chmod(job_id+"_run_refine.sh" , 0o777)

        bash_ansurr = open(job_id+"_run_ansurr.sh","w")
        bash_ansurr.write("#!/bin/bash\n")
        bash_ansurr.write("source "+str(os.path.abspath(args.env))+"/bin/activate\n")  
        bash_ansurr.write("python3 "+str(ansurr_script)+" $1 $2 $3 $4"+'\n')
        bash_ansurr.close()
        os.chmod(job_id+"_run_ansurr.sh" , 0o777)

        bash_bme = open(job_id+"_run_bme.sh","w")
        bash_bme.write("#!/bin/bash\n")
        bash_bme.write("source "+str(os.path.abspath(args.env))+"/bin/activate\n")  
        bash_bme.write("python3 "+str(bme_script)+" $1 $2 $3 $4"+'\n')
        bash_bme.close()
        os.chmod(job_id+"_run_bme.sh" , 0o777)

        dag = open(job_id+".dag","w")
        dag.write("JOB A "+job_id+"_run_refine.sub\n")
        dag.write("JOB B "+job_id+"_run_combine.sub\n")
        dag.write("JOB C "+job_id+"_run_ansurr.sub\n")
        dag.write("JOB D "+job_id+"_run_bme.sub\n")
        dag.write("PARENT A CHILD B\n")
        dag.write("PARENT B CHILD C\n")
        dag.write("PARENT C CHILD D\n")
        dag.close()

        check_quiet_print(args.quiet,home_dir+'/'+job_id,"submitted subsequent processing to the job queue, check NERRDS.log for progress")

        run_refine = subprocess.run(["condor_submit_dag", "-f", job_id+".dag"],stdout=subprocess.PIPE, stderr=open(os.devnull, 'w'))

    else: # run on this machine in serial 

        # generate conformers
        gen_conf.gen_conf(path_to_pdb, home_dir, job_id, pdb_id, conformers, average_rmsd)

        # refine conformers in serial
        for pdb in sorted(glob.glob(home_dir+'/'+job_id+"/anm/anm_"+pdb_id+"_*.pdb"),key=natsort):
            refine.refine(home_dir, job_id, pdb, cpu, str(tleap), str(sander), str(sander_MPI), str(cpptraj))

        # combine succesfully refined conformers into single ensemble
        refine.combine(home_dir, job_id, pdb_id)

        # run ansurr
        run_ansurr.run_ansurr(home_dir,job_id,pdb_id,path_to_shifts)

        # run BME
        reweight_bme.reweight_ensemble(home_dir,job_id,pdb_id,shifts_id)

if __name__ == "__main__":
    main()



import sys
import os
import glob
import shutil
import numpy as np

from importlib import resources
from pdbtools import pdb_mkensemble

with resources.path("nerrds.scripts.BME") as f:
	BME_path = f
sys.path.insert(1, BME_path)
import BME

natsort = lambda s: [int(t) if t.isdigit() else t.lower() for t in re.split('(\\d+)', s)] # used throughout to loop through files in natural order

def reweight(home_dir,job_id,job_id,shifts_id):
	bme_dir = home_dir+"/"+job_id+"/bme/"
	os.makedirs(bme_dir,exist_ok=True)
	os.chdir(bme_dir)

	# collate ansurr output into data files for BME
	data = glob.glob(home_dir+"/"+job_id+"/ansurr/anm_"+pdb_id+"_refined_"+shifts_id+"/ANSURR_output/out/*.out")

	calc = []
	exp = []
	labels = []
	n=0
	for i in sorted(data,key=natsort):
	    calc_tmp = []
	    for line in open(i,'r'):
	        if 'nan' not in line:
	            line = line.split()
	            if n==0:
	                exp.append([float(line[2]),0.05])
	                labels.append(line[0])   # do we need labels? probably yes because includes residue IDs actually used 
	                calc_tmp.append(float(line[3]))
	            else:
	                calc_tmp.append(float(line[3]))
	                
	    calc.append(calc_tmp)        
	    n=1

	calc = np.array(calc)
	exp = np.array(exp)  

	out = open(bme_dir+"exp.txt",'w')
	# BME wants to know what kind of the data is being used so it knows how to average the data. JCOUPLINGS is just a simple mean - fine for comparing per residue flexibility
	out.write("# DATA=JCOUPLINGS\n")  
	i=0
	for e in exp:
	    i+=1
	    out.write(str(i)+' '+str(e[0])+' '+str(e[1])+'\n')
	out.close()

	j=0
	out = open(bme_dir+"calc.txt",'w')
	for e in calc:
	    j+=1
	    out.write(str(j)+' ')
	    for i in e:
	        out.write(str(i)+' ')
	    out.write('\n')
	out.close()

	# define input file names
	exp_file = bme_dir+"exp.txt"
	calc_file = bme_dir+"calc.txt"

	# initialize. A name must be specified 
	rew = BME.Reweight("reweight_ANM")

	# load the experimental and calculated datasets
	rew.load(exp_file,calc_file)

	# 5-fold cross validation for 20 theta values
	thetas = np.geomspace(1,10000,20) # will this hold for everything?
	opt_theta = rew.theta_scan(thetas=thetas,nfold=5)

	# initialize. A name must be specified 
	rew = BME.Reweight("reweight_ANM_opt")

	# load the experimental and calculated datasets
	rew.load(exp_file,calc_file)

	# fit the data 
	chi2_before, chi2_after, phi = rew.fit(theta=opt_theta)

	# get weights
	weights = rew.get_weights()

	# copy chosen, named accoring to weight, and combined into a single file
	ensemble_dir = home_dir+"/"+job_id+"/ensemble/"
	os.makedirs(ensemble_dir,exist_ok=True)

	selected_models = [] # used later for plotting PCA
	selected_weights = [] # used later for plotting PCA

	mkensemble_string = []
	for i in enumerate(weights):
	    if i[1] > 0.01:   # have a cut off of needed to have at least 1% weight
	        print(i[0]+1,round(i[1],2),model_link[str(i[0]+1)])
	        shutil.copyfile(home_dir+"/"+job_id+"/refined/"+model_link[str(i[0]+1)].split('_refined.pdb')[0]+"/"+model_link[str(i[0]+1)],ensemble_dir+os.path.splitext(model_link[str(i[0]+1)])[0]+'_'+str(int(100*round(i[1],2)))+'.pdb')
	        mkensemble_string.append(ensemble_dir+os.path.splitext(model_link[str(i[0]+1)])[0]+'_'+str(int(100*round(i[1],2)))+'.pdb')
	        selected_models.append(i[0]+1)
	        selected_weights.append(round(i[1],2))
	        
	new_pdb = pdb_mkensemble.run(mkensemble_string)

	out = open(ensemble_dir+pdb_id+'_ensemble.pdb','w')
	for line in enumerate(new_pdb):
	    out.write(line[1])
	out.close()


import os
import sys
import shutil
import glob
import re

from pdbtools import pdb_splitmodel
from prody import *

from nerrds.functions import check_quiet_print

def gen_conf(path_to_pdb, home_dir, job_id, pdb_id, conformers, average_rmsd, quiet=False):

    natsort = lambda s: [int(t) if t.isdigit() else t.lower() for t in re.split('(\\d+)', s)] 

    #------------------------------------------------------ extract models from input ensemble --------------------------------------------------------------

    # split input PDB file into seperate models
    pdb_splitmodel.run(pdb_splitmodel.check_input([path_to_pdb]))

    # make dir to store split PDBs
    os.makedirs(job_id+'/models/',exist_ok=True)
    for pdb in glob.glob(pdb_id+'_*.pdb'):
        shutil.move(pdb,job_id+'/models/')

    num_models = len(glob.glob(home_dir+'/'+job_id+"/models/"+pdb_id+"_*.pdb"))

    check_quiet_print(quiet,home_dir+'/'+job_id,"generating "+str(conformers)+" conformers for each of "+str(num_models)+" models, ("+str(int(conformers) * int(num_models))+" in total)\nprody output:\n")

    #----------------------------------------------------------- generate conformers -------------------------------------------------------------------------

    # use prody run ANM to generate conformers for each model in input ensemble

    for pdb in sorted(glob.glob(home_dir+'/'+job_id+"/models/"+pdb_id+"_*.pdb"),key=natsort):
        
        parsed_pdb = parsePDB(pdb)
        parsed_pdb_ca = parsed_pdb.select('calpha')

        anm = ANM(pdb)
        anm.buildHessian(parsed_pdb_ca)
        anm.calcModes()

        bb_anm, all_atoms = extendModel(anm, parsed_pdb_ca, parsed_pdb.select('all'))

        ensemble = sampleModes(bb_anm[:3], all_atoms, n_confs=conformers, rmsd=average_rmsd)
        all_atoms = all_atoms.copy()
        all_atoms.addCoordset(ensemble)
        
        heavy = all_atoms.select('heavy')   # don't keep hydrogens, let AMBER add these during refinement to avoid naming issues

        os.makedirs(job_id+'/anm/',exist_ok=True) # save confomers here
        writePDB(job_id+'/anm/anm_'+os.path.basename(pdb), heavy)
        
        os.chdir(job_id+'/anm/')
        pdb_splitmodel.run(pdb_splitmodel.check_input(['anm_'+os.path.basename(pdb)]))
        os.remove('anm_'+os.path.basename(pdb))
        os.remove(glob.glob('anm_*_1.pdb')[0]) # remove the first model as this will be a submitted model - maybe we want to keep it?
        os.chdir(home_dir)

    # check conformer generation success
    generated_conformers = len(glob.glob(home_dir+"/"+job_id+"/anm/anm*pdb"))
    if generated_conformers > 0:
        check_quiet_print(quiet,home_dir+'/'+job_id,"\ngenerated "+str(generated_conformers)+" conformers")
    else:
        check_quiet_print(quiet,home_dir+'/'+job_id,"ERROR failed to generate any conformers, quitting")
        sys.exit(0)


if __name__ == "__main__":
    import sys

    path_to_pdb = sys.argv[1]
    home_dir = sys.argv[2]
    job_id = sys.argv[3]
    pdb_id = sys.argv[4]
    conformers = sys.argv[5]
    average_rmsd = sys.argv[6]

    gen_conf(path_to_pdb, home_dir, job_id, pdb_id, conformers, average_rmsd)


import os
import glob
import shutil

from ansurr import ansurr
from nerrds.functions import check_quiet_print

def run_ansurr(home_dir, job_id, pdb_id, path_to_shifts,quiet=False):

    check_quiet_print(quiet,home_dir+'/'+job_id,"running ANSURR")
    ansurr_dir = home_dir+"/"+job_id+"/ansurr/"
    os.makedirs(ansurr_dir,exist_ok=True)
    os.chdir(ansurr_dir)
    ansurr.main("-p "+home_dir+"/"+job_id+"/refined/anm_"+pdb_id+"_refined.pdb"+" -s "+path_to_shifts+" -r -q")

    ansurr_output = len(glob.glob(home_dir+"/"+job_id+"/ansurr/anm*/ANSURR_output/out/*.out"))

    to_remove = glob.glob(home_dir+"/"+job_id+"/ansurr/anm*/other_output/*")
    for t in to_remove:
        shutil.rmtree(t)

    # check ansurr success
    if ansurr_output > 0:
        check_quiet_print(quiet,home_dir+'/'+job_id,"ANSURR ran for "+str(ansurr_output)+" models")
    else:
        check_quiet_print(quiet,home_dir+'/'+job_id,"ERROR ANSURR failed to run for any models, quitting")
        sys.exit(0)

    shutil.rmtree(home_dir+"/"+job_id+"/ansurr/anm*/ANSURR_output/figs",ignore_errors=True)

if __name__ == "__main__":
    import sys
    home_dir = sys.argv[1]
    job_id = sys.argv[2]
    pdb_id = sys.argv[3]
    path_to_shifts = sys.argv[4]
    run_ansurr(home_dir, job_id, pdb_id, path_to_shifts)

